#pragma once

/*
Copyright (c) 2013-2016 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <stddef.h>
#include <uchar.h>
#include <stdint.h>
#include <inttypes.h>
#include <assert.h>

#include "config.h"

/** @file
 * @defgroup Definitions Definitions
 * @addtogroup Definitions
 * @{
 * @brief Declarations of the types and macros that are used extensively
 * throughout libssts.
 */

/**
 * @def S_BEGIN_DECLS
 * For interoperability with C++ we have to have this.
 *
 * Put at the beginning of header files.
 * @see S_END_DECLS
 */

/**
 * @def S_END_DECLS
 *
 * Put at end of header files.
 * @see S_BEGIN_DECLS
 */

#ifdef __cplusplus
#define S_BEGIN_DECLS extern "C" {
#define S_END_DECLS }
#else
#define S_BEGIN_DECLS
#define S_END_DECLS
#endif /*__cplusplus*/

S_BEGIN_DECLS

/* Make sure that we use utf-32 with the suchar string. */
#define STDC_UTF_32 1

/* ***************************** GENERIC MACROS ***************************** */

/** @def
 */

/**
 * @def S_DEPRECATED
 * Mark function as deprecated.
 */
/**
 * @def S_UNUSED
 * Supress compiler warning that variable is unused.
 */
/**
 * @def S_ALIGNED_8
 * Macro to align memory to 8 bit.
 */
/**
 * @def S_ALIGNED_16
 * Macro to align memory to 16 bit.
 */
#if defined(__GNUC__)
#define S_DEPRECATED __attribute__((deprecated))
#define S_UNUSED __attribute__((unused))
#define S_ALIGNED_8 __attribute__((aligned (8)))
#define S_ALIGNED_16 __attribute__((aligned (16)))
#elif defined(__MSC_VER)
#define S_DEPRECATED __declspec(deprecated)
// TODO, check if this rely works in MSVC.
#define S_UNUSED __pragma(warning(suppress:4100))
#define S_ALIGNED_8 __declspec(align(8))
#define S_ALIGNED_16 __declspec(align(16))
#else
#define S_DEPRECATED
#define S_UNUSED
#define S_ALIGNED_8
#define S_ALIGNED_16
#endif

/**
 * @def __S_ATOMIC__
 * The stupid atomic value.
 */
#if __STDC_NO_ATOMICS__
#define __S_ATOMIC__ volatile
#else
#define __S_ATOMIC__ __Atomic
#endif

/**
 * @def S_EXPORTED
 * Mark function as exported. This makes the function available from outside a
 * library on some platforms.
 *
 * @see S_HIDDEN
 */
/**
 * @def S_HIDDEN
 * Mark function as hidden. This hides a function so it is not exposed to the
 * outside of the library.
 *
 * @see S_EXPORTED
 */

/*
 * Copied from: https://gcc.gnu.org/wiki/Visibility .. with modifications.
 */
#if defined _WIN32 || defined __CYGWIN__
  #ifdef BUILDING_DLL
    #ifdef __GNUC__
      #define S_EXPORTED __attribute__ ((dllexport))
    #else
      #define S_EXPORTED __declspec(dllexport) // Note: actually gcc seems to also supports this syntax.
    #endif
  #else
    #ifdef __GNUC__
      #define S_EXPORTED __attribute__ ((dllimport))
    #else
      #define S_EXPORTED __declspec(dllimport) // Note: actually gcc seems to also supports this syntax.
    #endif
  #endif
  #define S_HIDDEN
#else
  #if __GNUC__ >= 4
    #define S_EXPORTED __attribute__ ((visibility ("default")))
    #define S_HIDDEN  __attribute__ ((visibility ("hidden")))
  #else
    #define S_EXPORTED
    #define S_HIDDEN
  #endif
#endif

/**
 * FALSE has the absolute value of 0, it is used in as a way to represent a
 * false value. A "not" value.
 *
 * @see sboolean
 */
#define FALSE 0

 /**
  * TRUE represents a value that is true. A value that is "not false".
  *
  * @see sboolean
  */
#define TRUE !FALSE

/**
 * Stupid macro to create a hash of a pointer.
 */
#define S_POINTER_TO_HASH_T(p) ((hash_t)(unsigned long) p)

/**
 * @defgroup Types Type declarations
 * @addtogroup Types
 * @{
 * @brief The typedefs that make up the types that are used in libssts.
 * 
 */

/**
 * sboolean is the standard definition of a boolean value in SSTS.
 *
 * @sa TRUE
 * @sa FALSE
 *
 * @note
 * This is the way it is done in GLib, so it is done here too.
 */
typedef int32_t sboolean;

/** hash type  */
typedef size_t hash_t;

/** spointer is a convinience typedef of void * */
typedef void* spointer;

/** sconstpointer is a convinience typedef of const void * */
typedef const void* sconstpointer;

/**
 * To prevent intercomparability problems we def schar and assert its size.
 */
typedef char schar;
static_assert (sizeof(schar) == sizeof (int8_t),
              "schar has the wrong size. Is not 8 bit long.");

/**
 * uchar is the representation for use with wide unicode strings and what not.
 *
 * The rationale for using char32_t as the unicode wide char instead of
 * wchar_t is that wchar_t's size depends on the implementation, compiler,
 * and system.
 *
 * To make the wide strings actually work as they should, even over network or
 * other communication we set this as the standard for the uchar.
 *
 * @warning This is not an unsigned char. If you need to store such small
 * values use #subyte
 */
typedef char32_t suchar;

/**
 * To prevent interchangeability problems we def long as int64_t.
 */
typedef int64_t slong;

/**
 * To prevent interchangeability problems we def int as int32_t.
 */
typedef int32_t sint;

/**
 * To prevent interchangeability problems we def short as int16_t.
 */
typedef int16_t sshort;

/**
 * To prevent interchangeability problems we def byte as int8_t.
 */
typedef int8_t sbyte;

/* *************************************
   ******* UNSIGNED VERSIONS ***********
   ************************************* */

/**
 * To prevent interchangeability problems we def ulong as uint64_t.
 */
typedef uint64_t sulong;

/**
 * To prevent interchangeability problems we def uint as uint32_t.
 */
typedef uint32_t suint;

/**
 * To prevent interchangeability problems we def ushort as uint16_t.
 */
typedef uint16_t sushort;

/**
 * To prevent interchangeability problems we def ubyte as uint8_t.
 */
typedef uint8_t subyte;

/* *********** FLOATS **************** */

typedef float sfloat;
static_assert (sizeof (sfloat) == sizeof (int8_t) * 4,
              "sfloat has the wrong size. Is not 32 bit.");


typedef double sdouble;
static_assert (sizeof (sdouble) == sizeof (int8_t) * 8,
              "sdouble has the wrong size. Is not 64 bit.");


typedef __float128 squadruple;
static_assert (sizeof (squadruple) == sizeof (int8_t) * 16,
              "squadruple has the wrong size. Is not 128 bit.");

/**
 * @}
 */

/**
 * Represents different types of objects.
 *
 * @se STypeName
 */
typedef enum {
  S_TYPE_NONE = 0, /**< Not a type. */
  S_TYPE_INT, /**< @ref sint */
  S_TYPE_LONG, /**< @ref slong */
  S_TYPE_SHORT, /**< @ref sshort */
  S_TYPE_CHAR, /**< @ref schar */
  S_TYPE_WCHAR, /**< @deprecated wide char (wchar_t) */
  S_TYPE_UCHAR, /**< @ref suchar */
  S_TYPE_UINT, /**< @ref suint */
  S_TYPE_ULONG, /**< @ref sulong */
  S_TYPE_USHORT, /**< @ref sushort */
  S_TYPE_BOOLEAN, /**< @ref sboolean */
  S_TYPE_STRING, /**< a string of @ref schar s */
  S_TYPE_WSTRING, /**< @deprecated  Platform specific wchar_t string */
  S_TYPE_USTRING, /**< a string of @ref suchar s */
  S_TYPE_UNUSED_0, /**< */
  S_TYPE_UNUSED_1, /**< */
  S_TYPE_UNUSED_2, /**< */
  S_TYPE_POINTER, /**< @ref spointer */
  S_TYPE_OBJECT, /**< @ref SObject */
  S_TYPE_INVALID /** Invalid type. */
} SType;

/** @brief
 * The names of the SType's
 *
 *
 */
static char * STypeName[] S_UNUSED = {
  "NONE",
  "INT",
  "LONG",
  "SHORT",
  "CHAR",
  "WCHAR",
  "UCHAR", /*< char32_t */
  "UINT",
  "ULONG",
  "USHORT"
  "BOOLEAN",
  "STRING",
  "WSTRING",
  "USTRING",
  "UNUSED_0/INVALID",
  "UNUSED_1/INVALID",
  "UNUSED_2/INVALID",
  "POINTER",
  "OBJECT",
  "INVALID",
  0x0,
  0x0,
};

/**
 * @brief Get the name of the SType.
 *
 * For use in bindings.
 */
char *
s_type_get_name (SType k);

/**
 * @defgroup Colours Colours
 * @addtogroup Colours
 * @{
 * @brief Printing colours for terminal/console printing.
 * 
 * Here is a list of macros for use when printing to the console.
 * @ref S_COLOR_RESET is used to reset the colours back to their original
 *      colour.
 * 
 * An example on how to use them can be found in utils.h:
 * @code{.c}
#define s_err_print(p, ...)\
  fprintf (stderr, S_COLOR_RED "[ERR] " p S_COLOR_RESET "\n", ##__VA_ARGS__)
 * @endcode
 */
 
#define S_COLOR_RESET   "\033[0m"       /**< Reset */
#define S_COLOR_BLACK   "\033[30m"      /**< Black */
#define S_COLOR_RED     "\033[31m"      /**< Red */
#define S_COLOR_GREEN   "\033[32m"      /**< Green */
#define S_COLOR_YELLOW  "\033[33m"      /**< Yellow */
#define S_COLOR_BLUE    "\033[34m"      /**< Blue */
#define S_COLOR_MAGENTA "\033[35m"      /**< Magenta */
#define S_COLOR_CYAN    "\033[36m"      /**< Cyan */
#define S_COLOR_WHITE   "\033[37m"      /**< White */
#define S_COLOR_BOLDBLACK   "\033[1m\033[30m"      /**< Bold Black */
#define S_COLOR_BOLDRED     "\033[1m\033[31m"      /**< Bold Red */
#define S_COLOR_BOLDGREEN   "\033[1m\033[32m"      /**< Bold Green */
#define S_COLOR_BOLDYELLOW  "\033[1m\033[33m"      /**< Bold Yellow */
#define S_COLOR_BOLDBLUE    "\033[1m\033[34m"      /**< Bold Blue */
#define S_COLOR_BOLDMAGENTA "\033[1m\033[35m"      /**< Bold Magenta */
#define S_COLOR_BOLDCYAN    "\033[1m\033[36m"      /**< Bold Cyan */
#define S_COLOR_BOLDWHITE   "\033[1m\033[37m"      /**< Bold White */

/** @} */

/**
 * @}
 */

/* We include this last so it does not create any problems.
 * We want people only to have to include defs.h and not mm.h anyway.
 *
 * It should be considered a continuation of defs.h.
 */
#include "mm.h"

S_END_DECLS
