/*
Copyright (c) 2013-2014 Gustav Hartvigsson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#pragma once

#include "defs.h"
#include <stddef.h>
#include "config.h"
#include <stdio.h>
#include <uchar.h>
#include <wchar.h>


S_BEGIN_DECLS

/** @file
 * @defgroup Utils Utils
 * @addtogroup Utils
 * @{
 * A collect of utility functions.
 *
 */

/**
 * Creats a new C string with the correct length from a long string.
 * This may be a costly operation.
 *
 * The restulting string must be freed by caller.
 */
S_EXPORTED
char *
s_string_new (const char * s);

/**
 * Creats a new C string with the correct length using standard fprint style
 * format.
 *
 * The restulting string must be freed by caller.
 */
S_EXPORTED
schar *
s_string_new_fmt (const schar * format, ...);


/**
 * Same as s_string_new, but with a set length.
 *
 * The restulting string must be freed by caller
 */
S_EXPORTED
schar *
s_string_new_with_len (const schar * s, size_t len);

/**
 * Are the strings equal?
 * @returns TRUE if a and b are equal.
 * @returns FALSE if the strings are not equal.
 */
S_EXPORTED
sboolean
s_string_is_equal (const schar * a, const schar * b);


/**
 * Gets the length of a string.
 * Normally uses strlen under the hood.
 */
S_EXPORTED
size_t
s_string_len (const schar * str);

/**
 * Get the length of the uchar string.
 *
 * @warning This is the count of suchar's, not necessarily the number of
 * characters in the string.
 */
S_EXPORTED
size_t
s_ustring_len (const suchar * us);

/**
 * Convert a wide (32 bit per char) string to a byte string.
 *
 * @param us The uchar string to convert to a schar string.
 *
 * @note The name is a little misleading, this (should) work with
 *       Unicode strings, but we do not real checking of unicode validity,
 *       and we have no formal support for Unicode above the basics.
 */
S_EXPORTED
schar *
s_ustring_to_string (const suchar * us);


/**
 * Convert (8bit per char) schar string to a (32bit per char) suchar string.
 *
 * @param str The schar string to convert to a suchar string.
 */
S_EXPORTED
suchar *
s_string_to_ustring (const schar * str);


/**
 * @deprecated Usage of this function is discouraged. Consider using uchar
 *             (char32_t) strings instead, and s_ustring_to_string(), for
 *             your own sanity.
 */
S_EXPORTED
S_DEPRECATED
schar *
s_wstring_to_string (const wchar_t * ws);

/**
 * Returns a C string with the current time (as run).
 *
 * The restulting string must be freed by caller.
 */
S_EXPORTED
schar *
s_current_time (void);

#if 0
/* strdup is not ISO C, so we have to declare it somewhare, this should work
 * even if we do not impliment the function ourselfs.
 *
 * This just to supress a compiler warning.
 */
#define strdup(s) s_string_new(s)

#endif /* if 0 */

/** Just a standard print function. */
#define s_print(p, ...)\
  fprintf (stdout, p, ##__VA_ARGS__)

/**
 * Print a yellow warning to stderr.
 */
#define s_warn_print(p, ...)\
  fprintf (stderr, S_COLOR_BOLDYELLOW "[WARN] " p "\n" S_COLOR_RESET, ##__VA_ARGS__)

/**
 * prints a red message with the prefix [ERR]
 */
#define s_err_print(p, ...)\
  fprintf (stderr, S_COLOR_RED "[ERR] " p S_COLOR_RESET "\n", ##__VA_ARGS__)

#if DEBUG
/**
 * debug_print is a function that only prints if compiled with the debug
 * flag not unset.
 */
  #define s_dbg_print(M, ...)\
    fprintf (stdout, S_COLOR_YELLOW "[DEBUG][%s:%d] " M S_COLOR_RESET "\n", __FILE__, __LINE__, ##__VA_ARGS__)
#else
  #define s_dbg_print(M, ...)
#endif

/**
 * Round a up a number to a multiple of an other.
 *
 * @param num The number to round up.
 * @param multiple The multiple to round up to.
 *
 * See: http://stackoverflow.com/a/9194117
 */
#define round_up(num, multiple) (((num + multiple - 1) / multiple) * multiple)


/**
 * rotate value count steps to the left and rotate the overflow.
 *
 * @param value The value to rotate.
 * @param count The number of bits to shift it.
 *
 * @warning count must be positive.
 */
#define rotl(value, count)\
  ((value << count) | (value >> ((-count) & (sizeof(value) * 8) - 1)))

/**
 * rotate value count steps to the rights and rotate the overflow.
 *
 * @param value The value to rotate.
 * @param count The number of bits to shift it.
 *
 * @warning count must be positive.
 */
#define rotr(value, count)\
  ((value >> count) | (value << ((-count) & (sizeof(value) * 8) - 1)))


/* -------- TraceBack stuff ---- */

#define S_STACK_TRACKE_SIZE 128

#ifdef  __unix__
  //#pragma message ("We are a UNIX.")
  #include <execinfo.h>
  /**
   * Get a platform specific traceback.
   *
   * Works on UNIX's and Windows (Not working).
   */
  #define print_backtrace() {\
    fprintf (stderr, "[BACKTRACE:]\n");\
    void ** _backtrace_data_ = s_calloc (S_STACK_TRACKE_SIZE, sizeof (void *));\
    int _backtrace_len_ = backtrace (_backtrace_data_, 10);\
    char ** _backtrace_strs_ = backtrace_symbols (_backtrace_data_, _backtrace_len_);\
    if (_backtrace_strs_ == NULL) {\
      fprintf (stderr, "Could not get backtrace...\n");\
    } else {\
      for (int i = 0; i < _backtrace_len_; i++) {\
        fprintf (stderr, "%s\n", _backtrace_strs_[i]);\
      }\
      s_free (_backtrace_strs_);\
    }\
    fprintf (stderr, "[END BACKTRACE]\n");\
  }
#elif __WIN32__ || __WIN64__
  //#pragma message ("We are a Windows (why, oh why?)")
  #include <windows.h>
  #include <dbghelp.h>
  #include <winbase.h>
  #if __WIN64__
    #define _PLATFORM_SPECIFIC_DWORD DWORD64
  #else
    #define _PLATFORM_SPECIFIC_DWORD DWORD
  #endif
  #define print_backtrace() {\
    fprintf (stderr, "[BACKTRACE:]\n");\
    void * _backtrace_stack[S_STACK_TRACKE_SIZE];\
    HANDLE _backtrace_proc = GetCurrentProcess ();\
    SymInitialize (_backtrace_proc, 0, TRUE);\
    unsigned short _backtrace_frames = CaptureStackBackTrace ( 0, S_STACK_TRACKE_SIZE, _backtrace_stack, NULL );\
    SYMBOL_INFO * _backtrace_symbol = s_calloc (1, sizeof (SYMBOL_INFOW) + (sizeof (char) * 256));\
    _backtrace_symbol->MaxNameLen = 255;\
    _backtrace_symbol->SizeOfStruct = sizeof (SYMBOL_INFO);\
    for (int i = 0; i <= _backtrace_frames; i++) {\
      SymFromAddr (_backtrace_proc, (_PLATFORM_SPECIFIC_DWORD)(_backtrace_stack[i]), 0, _backtrace_symbol);\
      fprintf (stderr, "%i: %s - 0x%0X\n", _backtrace_frames - i, _backtrace_symbol->Name, _backtrace_symbol->Address);\
    }\
    SymCleanup (_backtrace_proc);\
    s_free (_backtrace_symbol);\
    fprintf (stderr, "[END BACKTRACE]\n");\
  }
#else
  //#pragma message ("We are not a UNIX or a Windows")
  #define print_backtrace() {\
    fprintf(stdout, "[Can not get backtrace]\nCurrent Function:" __func__);\
    fprintf (stderr, "[END BACKTRACE]\n");\
  }
#endif


/** @} */

S_BEGIN_DECLS
